import sbol3
import networkx as nx
from .geneproduct import Regulator, Reporter
from .supplement import Supplement
from .operators.not_ import Not
from .operators.nor import Nor
from .operators.buffer import Buffer
from .operators.receiver import Receiver
from .operators.source import Source
from typing import List #, Dict, Tuple, Optional, Union, Any

class GeneticNetwork():
    """
    Representation of a genetic netowrk composed by a set of Operators, Regulators and Reporters.

    ...

    Attributes
    ----------
    operators : List[Operator]
        List of Operators that are part of the genetic network
    regulators : List[Regulator]
        List of Regulators that are part of the genetic network
    reporters : List[Reporter]
        List of Reporters that are part of the genetic network
    vector : int
        Flapjack ID of the vector that is associated with the genetic network

    Methods
    -------
    to_graph()
        Builds a graph representation of the genetic netwok
    draw()
        Generates a plot of the graph representation builded by to_graph()
    to_sbol(sbol_doc=None)
        Generates a SBOL3 Document representation of the genetic network on sbol_doc  
    """
    def __init__(self, vector=None):
        self.operators = []
        self.regulators = []
        self.reporters = []
        self.vector = vector

    def initialize(self):
        for regulator in self.regulators:
            regulator.initialize()
        for reporter in self.reporters:
            reporter.initialize()

    def add_operator(self, op):
        self.operators.append(op)

    def add_regulator(self, reg):
        self.regulators.append(reg)

    def add_reporter(self, rep):
        self.reporters.append(rep)

    def add_operators(self, ops):
        for op in ops:
            self.operators.append(op)

    def add_regulators(self, regs):
        for reg in regs:
            self.regulators.append(reg)

    def add_reporters(self, reps):
        for rep in reps:
            self.reporters.append(rep)

    def step(self, growth_rate=1, t=0, dt=0.1):
        for op in self.operators:
            expression_rate = op.expression_rate(t, dt)
            op.output.express(expression_rate)

        for regulator in self.regulators:
            regulator.step(growth_rate, dt)

        for reporter in self.reporters:
            reporter.step(growth_rate, dt)

    def to_graph(self):
        g = nx.DiGraph()
        for op in self.operators:
            if hasattr(op, 'input'):
                if type(op.input)==list:
                    for i in op.input:
                        g.add_edge(i, op)
                else:
                    g.add_edge(op.input, op)
            g.add_edge(op, op.output)
        return g

    def draw(
        self,
        node_shape='o',
        node_size=200,
        linewidths=2,
        alpha=0.75,
        arrowsize=5,
        font_size=6,
        font_family='Tahoma',
        font_weight='bold',
        pos=nx.kamada_kawai_layout
        ):
        g = self.to_graph()
        pos = pos(g)
        nx.draw_networkx_nodes(
            g, 
            pos=pos, 
            node_color=[n.color for n in g.nodes], 
            node_shape=node_shape, 
            node_size=node_size,
            linewidths=linewidths,
            alpha=alpha
            )
        nx.draw_networkx_edges(
            g, 
            pos=pos, 
            width=1, 
            node_shape=node_shape, 
            node_size=node_size,
            arrowsize=arrowsize
            )
        nx.draw_networkx_labels(
            g,
            pos=pos,
            font_size=font_size,
            font_family=font_family,
            font_weight=font_weight
            )

    def to_sbol(self, sbol_doc: sbol3.Document = None) -> sbol3.Document:
        """Convert the genetic network to SBOL.
        :param sbol_doc: The SBOL document to add the genetic network to.
        """
        if sbol_doc:
            doc=sbol_doc
        else: 
            print('No SBOL Document provided')
            print('Generating a new SBOL Document')
            doc = sbol3.Document()
        products = set()
        geneticnetwork = sbol3.Component('geneticnetwork', sbol3.SBO_DNA)
        geneticnetwork.roles.append(sbol3.SO_ENGINEERED_REGION)
        loica_set = set()
        for op in self.operators:
            operator_comp = op.sbol_comp
            output_comp = op.output.sbol_comp
            operator_sc = sbol3.SubComponent(operator_comp)
            output_sc = sbol3.SubComponent(output_comp)
            # TODO output string for policistronic operators
            if type(op)==Source:
                input_str= 'c'
                tu = sbol3.Component(f'TU_{op}_{op.output.name}', sbol3.SBO_DNA) #generalize to multi input/output TUs
                tu.roles.append(sbol3.SO_ENGINEERED_REGION)
                tu.features = [operator_sc, output_sc] 
            elif type(op)==Nor: # type(op.input)==List:
                input_str = ''
                tu = sbol3.Component(f'TU{input_str}_{op}_{op.output.name}', sbol3.SBO_DNA) #generalize to multi input/output TUs
                tu.features = [operator_sc, output_sc]
                for inp in op.input:
                    input_str += f'_{inp.name}'
                    input_comp = inp.sbol_comp
                    if type(input_comp)==sbol3.Component:
                        input_sc = sbol3.SubComponent(input_comp)
                        tu.features.append(input_sc)
                    else:
                        tu.features.append(input_comp)
            else:
                input_str= f'_{op.input.name}'
                tu = sbol3.Component(f'TU{input_str}_{op}_{op.output.name}', sbol3.SBO_DNA) #generalize to multi input/output TUs
                tu.features = [operator_sc, output_sc]
                input_comp = op.input.sbol_comp
                if type(input_comp)==sbol3.Component:
                    input_sc = sbol3.SubComponent(input_comp)
                    tu.features.append(input_sc)
                else:
                    tu.features.append(input_comp)                  

            # TU Component
            tu.roles.append(sbol3.SO_ENGINEERED_REGION)              
            tu.constraints = [sbol3.Constraint(sbol3.SBOL_PRECEDES, operator_sc, output_sc)]
            # generate a sequence for the TU assuming assembly by type IIS REsnf both parts will have the fusion sites.
            # compare last 4 bp with thefirst 4 bp of the next part, given the preceds constraint.
            # if they are the same then delete one of them and concatenate the rest.
            # else error or comment TU sequence can not be generated, provide ways to add it.

            # Output GeneProduct Component
            if op.output != List:
                outputs = [op.output]
            else: outputs = op.output
            for op_output in outputs:
                if type(op_output)==Regulator:
                    if op_output.type_ == 'PRO':
                        output_gp_comp = sbol3.Component(f'{op.output.name}_protein', sbol3.SBO_PROTEIN)
                        output_gp_comp.roles.append(sbol3.SO_TRANSCRIPTION_FACTOR)               
                    elif op_output.type_ == 'RNA':
                        output_gp_comp = sbol3.Component(f'{op.output.name}_rna', sbol3.SBO_RNA)
                        output_gp_comp.roles.append(sbol3.SO_TRANSCRIPTION_FACTOR)
                    else: 
                        print('Unsupported output molecule type')
                elif type(op_output)==Reporter: # For now just support fluorescent reporters
                    if op_output.type_ == 'PRO':
                        output_gp_comp = sbol3.Component(f'{op.output.name}_protein', sbol3.SBO_PROTEIN)
                        output_gp_comp.roles.append('http://purl.obolibrary.org/obo/NCIT_C37894')
                    elif op_output.type_ == 'RNA':
                        output_gp_comp = sbol3.Component(f'{op.output.name}_rna', sbol3.SBO_RNA)
                        output_gp_comp.roles.append('http://purl.obolibrary.org/obo/NCIT_C37894')  
                    else: 
                            print('Unsupported output molecule type')
                else:
                    print('Unsupported output Type')
                output_gp_sc = sbol3.SubComponent(output_gp_comp)
                tu.features.append(output_gp_sc)
                if op_output not in products:
                    products.add(op_output) 
                    loica_set.add(output_gp_comp)
                # Genetic Production Interaction pf the output
                output_participation = sbol3.Participation(roles=[sbol3.SBO_TEMPLATE], participant=output_sc)
                gp_participation = sbol3.Participation(roles=[sbol3.SBO_PRODUCT], participant=output_gp_sc)
                production = sbol3.Interaction(types=[sbol3.SBO_GENETIC_PRODUCTION], participations=[output_participation, gp_participation])
                tu.interactions.append(production)
            # obtain TU subcomponents sequences, specially CDS and flanking parts sequences
            # look for ATG on the CDS and upstream part sequences (in the case of MoClo the ATG is in the fusion sites)
            # look for stop codons on frame with the ATG.
            # add translated the sequence between the ATG and the stop codon as protein sequence.
            #protein.sequence = tu.cds.sequence

            # Input Product Component
            if type(op) == Source:
                inputs=[]
            elif type(op) == Nor: #type(op.input) != List:
                inputs = op.input
            else: inputs = [op.input]
            #inputs_prod_sc = []
            for op_input in inputs:
                if type(op_input)==Regulator:
                    if op_input.type_ == 'PRO':
                        input_prod_comp = sbol3.Component(f'{op_input.name}_protein', sbol3.SBO_PROTEIN)
                        input_prod_comp.roles.append(sbol3.SO_TRANSCRIPTION_FACTOR)               
                    elif op_input.type_ == 'RNA':
                        input_prod_comp = sbol3.Component(f'{op_input.name}_rna', sbol3.SBO_RNA)
                        input_prod_comp.roles.append(sbol3.SO_TRANSCRIPTION_FACTOR)
                    else: 
                        print('Unsupported input molecule type')
                elif type(op_input)==Supplement:
                    input_prod_comp = sbol3.Component(f'{op_input.name}_chemical', sbol3.SBO_SIMPLE_CHEMICAL)
                    input_prod_comp.roles.append(sbol3.SO_TRANSCRIPTION_FACTOR)     
                else:
                    print('Unsupported input Type')
                # adds two times prod comp on the repressilator but necessary for normal circuits
                if op_input not in products:
                    products.add(op_input) 
                    loica_set.add(input_prod_comp)
                
                input_prod_sc = sbol3.SubComponent(input_prod_comp)
                tu.features.append(input_prod_sc)
                #inputs_prod_sc.append(input_prod_sc)
                #how can I not create 2 times the same component?
                if type(op_input)!=Regulator: # if it is a regulator it is already created
                    loica_set.add(input_prod_comp)
                # Input Interaction
                if type(op)==Not or Nor:
                    input_participation = sbol3.Participation(roles=[sbol3.SBO_INHIBITOR], participant=input_prod_sc)
                    op_participation = sbol3.Participation(roles=[sbol3.SBO_INHIBITED], participant=operator_sc)
                    interaction = sbol3.Interaction(types=[sbol3.SBO_INHIBITION], participations=[input_participation, op_participation])
                    tu.interactions.append(interaction)
                elif type(op)==Buffer or Receiver:
                    input_participation = sbol3.Participation(roles=[sbol3.SBO_STIMULATOR], participant=input_prod_sc)
                    op_participation = sbol3.Participation(roles=[sbol3.SBO_STIMULATED], participant=operator_sc)
                    interaction = sbol3.Interaction(types=[sbol3.SBO_STIMULATION], participations=[input_participation, op_participation])
                    tu.interactions.append(interaction)
                elif type(op)==Source:
                    pass
                else:
                    print('Unsupported operator Type')         
            # Model
            #model_string = str(op.__dict__)
            op_model = sbol3.Model(f'LOICA{input_str}_{op}_{op.output.name}_model', 
                            source='https://github.com/SynBioUC/LOICA/blob/master/loica/operators',
                            language='http://identifiers.org/EDAM:format_3996',
                            framework='http://identifiers.org/SBO:0000062',)
                            #attachments=[model_string])
            doc.add(op_model)
            tu.models.append(op_model)
            doc.add(tu)
            tu_sc = sbol3.SubComponent(tu)
            geneticnetwork.features.append(tu_sc)
        loica_list = list(loica_set)
        doc.add(loica_list) 
        if len(geneticnetwork.features) > 1:
            for i in range(len(geneticnetwork.features)-1):
                geneticnetwork.constraints = [sbol3.Constraint(sbol3.SBOL_PRECEDES, geneticnetwork.features[i], geneticnetwork.features[i+1])]
        else: pass
        doc.add(geneticnetwork)
        return doc

        



import pandas as pd
import numpy as np
from tqdm import tqdm
from . import genetic_network, metabolism



class Assay:
    """
    Assay measures a set of samples in parallel at a set of timepoints.
    Connects to flapjack to generate data, and to fit parameters to data.
    
    ...
    
    Attributes
    ----------
    samples : List[Sample]
        List of Samples that belongs to the Assay
    n_measurements : int
        Number of measurements to take
    interval : int
        Time in hours between each measurements
    name : str
        Name of the Assay
    description: str
        Descriptioin of the Assay
    biomass_signal_id : int
        Flapjack ID of the Assay that is associated with the Assay

    Methods
    -------
    run(substeps=10, nsr=0, biomass_bg=0, fluo_bg=0)
        Runs the Assay time series
    upload(flapjack, study)
        Upload the data produced by running the Assay to Flapjack into the Study
    """
    def __init__(self, 
            samples, 
            n_measurements, 
            interval,
            name='Loica assay',
            description='',
            biomass_signal_id=None
            ):
        '''
        Assay measures a set of samples in parallel at a set of timepoints
        Connects to flapjack to generate data, and to fit parameters to data

        '''
        self.samples = samples
        self.n_measurements = n_measurements
        self.interval = interval
        self.measurements = pd.DataFrame()
        self.name = name
        self.description = description
        self.biomass_signal_id = biomass_signal_id

    def run(self, substeps=10, nsr=0, biomass_bg=0, fluo_bg=0):
        '''
        Run the assay measuring at specified time points, with simulation time step dt
        '''
        #substeps = self.interval / dt
        dt = self.interval / substeps
        n_samples = len(self.samples)
        with tqdm(total=100) as pbar:
            for sample_id, sample in enumerate(self.samples):
                sample.initialize()
                # Integrate models
                for t in range(self.n_measurements):
                    time = t * self.interval
                    # Record measurements of fluorescence
                    for reporter in sample.reporters:
                        sig = reporter.concentration
                        signal_id = reporter.signal_id
                        signal_name = reporter.name
                        noise = np.random.normal(scale=np.sqrt(nsr))
                        meas = sig * sample.biomass(time) + fluo_bg
                        noisy_meas = (1 + noise) * meas
                        noise_bg = np.random.normal(scale=np.sqrt(nsr))
                        corr_meas = noisy_meas - (1 + noise_bg) * fluo_bg
                        row = {
                                'Time': time, 
                                'Signal_id': signal_id, 
                                'Signal':signal_name, 
                                'Measurement': corr_meas,
                                'Sample':sample_id
                                }
                        self.measurements = self.measurements.append(row, ignore_index=True)
                    # Record measurement of biomass
                    noise = np.random.normal(scale=np.sqrt(nsr))
                    meas = sample.biomass(time) + biomass_bg
                    noisy_meas = (1 + noise) * meas
                    noise_bg = np.random.normal(scale=np.sqrt(nsr))
                    corr_meas = noisy_meas - (1 + noise_bg) * biomass_bg
                    row = {
                            'Time': time, 
                            'Signal_id': self.biomass_signal_id, 
                            'Measurement': corr_meas,
                            'Signal':'Biomass', 
                            'Sample':sample_id
                            }
                    self.measurements = self.measurements.append(row, ignore_index=True)
                    # Compute next time step
                    for tt in range(substeps):
                        time = t * self.interval + tt * dt
                        sample.step(time, dt)
                pbar.update(1 / n_samples * 100)
            pbar.close()
                
    def upload(self, flapjack, study):
        assay = flapjack.create('assay', 
                                    name=self.name, 
                                    study=study, 
                                    temperature=0, 
                                    machine='Loica',
                                    description=self.description)
        n_samples = len(self.samples)
        with tqdm(total=100) as pbar:
            for sample_id,sample in enumerate(self.samples):
                fj_sample = flapjack.create('sample',
                                    row=sample_id, col=1,
                                    media=sample.media,
                                    strain=sample.strain,
                                    vector=sample.vector,
                                    assay=assay.id[0],
                                    )

                supplements = []
                for supp,conc in sample.supplements.items():
                    chemical = flapjack.get('chemical', name=supp.name)
                    if len(chemical)==0:
                        chemical = flapjack.create('chemical', name=supp.name, description='Testing')
                    name = supp.name + f' {conc}'
                    s = flapjack.get('supplement', name=name, concentration=conc, chemical=chemical.id[0])
                    if len(s)==0:
                        s = flapjack.create('supplement', name=name, concentration=conc, chemical=chemical.id[0])
                    supplements.append(s.id[0])
                if len(supplements):
                    flapjack.patch('sample', fj_sample.id[0], supplements=supplements)

                meas = self.measurements[self.measurements.Sample==sample_id]
                for signal_id,data in meas.groupby('Signal_id'):
                    if signal_id:
                        flapjack.upload_measurements(data, signal=[signal_id], sample=fj_sample.id)
                pbar.update(1 / n_samples * 100)
            pbar.close()



import numpy as np

class Colony:
    def __init__(self, 
            circuit=None, 
            r0=1,
            mu0=1
            ):
        self.circuit = circuit
        self.r0 = r0
        self.mu0 = mu0

    def fun(self, x):
        # Compute growth rates on mesh
        mu = np.exp(-x)
        vel = 0.5 * (1 -np.exp(-x/self.r0))
        dx = np.mean(np.diff(x))
        geneprods =  self.circuit.reporters + self.circuit.regulators
        nprods = len(geneprods)
        
        def step(t, y):
            y = y.reshape((len(x),nprods))
            dy = np.zeros_like(y)

            for g,prod in enumerate(geneprods):
                prod.concentration = y[:,g]
                prod.expression_rate = 0

            for o,op in enumerate(self.circuit.operators):
                expression_rate = op.expression_rate(t, 0)
                op.output.express(expression_rate)

            for g,prod in enumerate(geneprods):
                pi = prod.concentration
                # Finite difference spatial derivatives
                dpidx = np.zeros_like(x)
                for i in range(1, len(x)-1):
                    dpidx[i] = (pi[i] - pi[i-1]) / dx
                dpidx[-1] = (pi[-1] - pi[-2]) / dx
                dpidx[0] = (pi[1] - pi[0]) / dx

                # Update protein concs
                dpidt =  prod.expression_rate - (prod.degradation_rate + mu*self.mu0) * pi - vel*dpidx
                dy[:,g] = dpidt

            return dy.ravel()
        return step

    def map_kymo(self, kymo):
        rkymo = np.zeros_like(kymo)
        nx,nt,_ = kymo.shape
        for t in range(nt):
            for xx in range(((t*nx)//nt)):
                rkymo[-xx+((t*nx)//nt),t,:] = kymo[xx,t,:]
        return rkymo

    def norm_kymo(self, kymo):
        nkymo = np.zeros_like(kymo)
        for c in range(nkymo.shape[2]):
            nkymo[:,:,c] = kymo[:,:,c] / kymo[:,:,c].max()
        return nkymo

    def kymograph(self, nx, t0, tmax):
        geneprods =  self.circuit.reporters + self.circuit.regulators
        nprods = len(geneprods)
        L = (tmax - t0) / 2
        x = np.linspace(0, L, nx)

        dx = np.diff(x).mean()
        dt = dx * 0.5
        nt = int((tmax-t0) // dt)
        dydt = self.fun(x)
        y = np.zeros((nx,nprods,nt))
        for g,prod in enumerate(geneprods):
            prod.initialize()
            y[:,g,0] = prod.init_concentration

        for t in range(1,nt):
            y[:,:,t] = y[:,:,t-1] + dydt(t*dt, y[:,:,t-1]).reshape((nx,nprods)) * dt

        nreps = len(self.circuit.reporters)
        kymo = np.zeros((nx,nt,nreps))
        for o in range(nreps):
            kymo[:,:,o] = y[:,o,:]
        return kymo




class Sample:
    """
    Representation of a sample that encapsulates GeneticNetwork and Metabolism.
    Incorporate environment information such as Supplements or chemicals, strain and media. 
    Ex: 1 well in a plate, single cell.
    ...

    Attributes
    ----------
    genetic_network : GeneticNetwork
        genetic network that is part of the sample
    metabolism : Metabolism
        metabolism that drives the genetic network in the sample
    assay : Assay
        assay to which this sample belongs
    media : str
        Name of the media in the sample
    strain : str
        Name of the strain in the sample
    
     Methods
    -------
    add_supplement(supplement, concentration)
        stablishes the concentration of Supplement
    """
    def __init__(self, 
            genetic_network=None, 
            metabolism=None, 
            assay=None,
            media=None,
            strain=None,
            ):
        self.genetic_network = genetic_network
        self.metabolism = metabolism
        self.media = media
        self.strain = strain
        self.vector = self.genetic_network.vector
        if self.genetic_network:
            self.reporters = self.genetic_network.reporters
        if metabolism:
            self.biomass = self.metabolism.biomass
        self.supplements = {}

    def initialize(self):
        self.genetic_network.initialize()

    def add_supplement(self, supplement, concentration):
        self.supplements[supplement] = concentration

    def step(self, t, dt):
        if self.genetic_network and self.metabolism:
            growth_rate = self.metabolism.growth_rate(t)
            for supp,conc in self.supplements.items():
                supp.concentration = conc
            self.genetic_network.step(growth_rate, t, dt)
            self.reporters = self.genetic_network.reporters





import numpy as np
from numpy.fft import fft, ifft, fftfreq
from scipy.optimize import least_squares
from scipy.interpolate import interp1d

from .source import *
from flapjack import *

class Receiver:
    """
    A class that represents a DNA fragment that encode a genetic operator.
    The Receiver Operator is an abstraction of an inducible promoter that
    maps an external input into an output using a Hill function.

    ...
    
    Attributes
    ----------
    input : Regulator | Supplement
        The input of the operator that regulates the expression of the output
    output : Regulator | Reporter
        The output of the operator that is regulated by the input
    a : int | float
        Basal expression rate, LOW
    b : int | float
        Regulated expression rate, HIGH
    K : int | float
        Half expression input concentration
    n : int | float
        Hill coefficient, cooperative degree
    uri : str, optional
        SynBioHub URI
    sbol_comp : SBOL Component, optional
        SBOL Component

    Methods
    -------
    characterize(flapjack, receiver, inverter, media, strain, signal, biomass_signal, gamma)
        Parameterize the Operator model that maps Input concentration into Output expression rate
    """
    color = 'orange'
    def __init__(self, input, output, a, b, K, n, uri=None, sbol_comp=None):
        self.a = a
        self.b = b
        self.K = K
        self.n = n
        self.input = input
        self.output = output
        self.uri = uri
        self.sbol_comp = sbol_comp

    def __str__(self):
        return 'REC'
        
    def expression_rate(self, t, dt):
        inducer = self.input.concentration
        i = (inducer/self.K)**self.n
        expression_rate = ( self.a + self.b*i ) / (1 + i)
        return expression_rate

    def forward_model(
            self,
            a=0,
            b=1,
            K_A=1,
            n_A=2,
            Dt=0.05,
            sim_steps=10,
            A=[0],
            odval=[1]*100,
            gamma=0,
            p0=0,
            nt=100
        ):
        p1_list,A_list,t_list = [],[],[]
        p1 = p0
        for t in range(nt):
            p1_list.append(p1)
            A_list.append(A)
            t_list.append(t * Dt)
            od = odval[t]
            tt = t*Dt
            for tt in range(sim_steps):
                aa = (A/K_A)**n_A
                nextp1 = p1 + (od * (a + b*aa)/(1 + aa) - gamma*p1) * Dt/sim_steps
                p1 = nextp1

        ap1 = np.array(p1_list).transpose()
        AA = np.array(A_list).transpose()
        tt = np.array(t_list).transpose()
        t = np.arange(nt) * Dt
        return ap1,AA,tt

    def residuals(self, df, oddf): 
        def func(x): 
            a = x[0]
            b = x[1]        
            K_A = x[2]
            n_A = x[3]        
            residual_list = []
            df_sorted = df.sort_values(['Sample', 'Time'])
            oddf_sorted = oddf.sort_values(['Sample', 'Time'])
            for samp_id,samp_data in df_sorted.groupby('Sample'):
                odval = oddf_sorted[oddf_sorted.Sample==samp_id].Measurement.values
                data = samp_data.Measurement.values
                p0 = data[0]
                A = samp_data.Concentration1.values[0]
                t = samp_data.Time.values
                dt = np.mean(np.diff(t))
                nt = len(t)
                p,AA,tt = self.forward_model(
                            a=a,
                            b=b,
                            K_A=K_A,
                            n_A=n_A,
                            Dt=dt,
                            A=A,
                            odval=odval,
                            nt=nt,
                            p0=p0
                        )
                model = p.ravel()
                residual = data - model
                residual_list.append(residual) 
            return np.array(residual_list).ravel()
        return func


    def characterize(self, flapjack, vector, media, strain, signal, biomass_signal):
        expression_df = flapjack.analysis(media=media, 
                            strain=strain,
                            vector=vector,
                            signal=signal,
                            type='Background Correct',
                            biomass_signal=biomass_signal
                            )

        biomass_df = flapjack.analysis(media=media, 
                            strain=strain,
                            vector=vector,
                            signal=biomass_signal,
                            type='Background Correct',
                            biomass_signal=biomass_signal
                            )

        # Bounds for fitting (a,b,K,n)
        lower_bounds = [0, 0, 0, 0]
        upper_bounds = [1e8, 1e8, 1e8, 6]
        bounds = [lower_bounds, upper_bounds]
        '''
            a = x[0]
            b = x[1]
            K_A = x[2]
            n_A = x[3]
        '''

        res = least_squares(
                self.residuals(
                    expression_df, biomass_df
                    ), 
                [0, 1, 1, 1], 
                bounds=bounds
                )
        self.res = res
        self.a = res.x[0]
        self.b = res.x[1]
        self.K = res.x[2]
        self.n = res.x[3]



class Buffer:
    """
    A class that represents a DNA fragment that encode a genetic operator.
    The Buffer Operator is an abstraction of an inducible promoter that
    maps an input into an output using a Hill function.

    ...
    
    Attributes
    ----------
    input : Regulator | Supplement
        The input of the operator that regulates the expression of the output
    output : Regulator | Reporter
        The output of the operator that is regulated by the input
    a : int | float
        Basal expression rate, LOW
    b : int | float
        Regulated expression rate, HIGH
    K : int | float
        Half expression input concentration
    n : int | float
        Hill coefficient, cooperative degree
    uri : str, optional
        SynBioHub URI
    sbol_comp : SBOL Component, optional
        SBOL Component

    Methods
    -------
    characterize(flapjack, receiver, inverter, media, strain, signal, biomass_signal, gamma)
        Parameterize the Operator model that maps Input concentration into Output expression rate
    """
    def __init__(self, input, output, a, b, K, uri=None, sbol_comp=None):
        self.a = a
        self.b = b
        self.K = K
        self.n = n
        self.input = input
        self.output = output
        self.uri = uri
        self.sbol_comp = sbol_comp
        
    def expression_rate(self, t, dt):
        input_inducer = self.input.concentration
        i = (input_inducer/self.K)**self.n
        expression_rate = ( self.a + self.b*i ) / (1 + i)
        return expression_rate





import numpy as np
from scipy.interpolate import interp1d

def ramp_growth_rate(t, start, slope):
    gr = np.maximum(0, slope*(t-start))
    return(gr)

def ramp_biomass(t, od0, start, slope):
    logod = np.maximum(0, ((t-start)**2)/2)
    od = od0 * np.exp(logod)
    return(od)

def step_growth_rate(t, start):
    gr = 1 * ((t-start)>0)
    return(gr)

def step_biomass(t, od0, start):
    logod = np.maximum(0, t-start)
    od = od0 * np.exp(logod)
    return(od)

def gompertz_growth_rate(t, y0, ymax, um, l):
    A = np.log(ymax/y0)
    gr = um *np.exp((np.exp(1)* um *(l - t))/A - \
            np.exp((np.exp(1)* um *(l - t))/A + 1) + 2)
    return(gr)

def gompertz(t, y0, ymax, um, l):
    A = np.log(ymax/y0)
    log_rel_od = (A*np.exp(-np.exp((((um*np.exp(1))/A)*(l-t))+1)))
    od = y0 * np.exp(log_rel_od)
    return(od)

class Metabolism:
    """
    Context for gene expression, incorporates biomass and growth rate.

    """
    def __init__(self):
        pass

class SimulatedMetabolism(Metabolism):
    """
    Simulated context for gene expression, incorporates biomass and growth rate.
    ...

    Attributes
    ----------
    biomass
        A function of time that describes biomass f(t)=biomass
    growth_rate
        A function of time that describes the growth rate f(t)=growth rate
    """
    def __init__(self, biomass_func, growth_rate_func):
        super().__init__()
        self.biomass = biomass_func
        self.growth_rate = growth_rate_func

class DataMetabolism(Metabolism):
    """
    Characterized context for gene expression, incorporates biomass and growth rate.
    ...

    Attributes
    ----------
    fj : Flapjack
        Flapjack instance used to fetch data from
    media : str
        Name of the media to query
    strain : str
        Name of the strain to query
    vector : str
        Name of the vector to query
    biomass_signal : str
        Name of signal to query and use as biomass
    
    Methods
    -------
    biomass(t)
        Return biomass at a given time from characterization data
    growth:rate(t)
        Return growth rate at a given time from characterization data
    """
    def __init__(self, fj, media, strain, vector, biomass_signal):
        super().__init__()
        gr = fj.analysis(media=media.id, 
                    strain=strain.id,
                    vector=vector.id,
                    type='Expression Rate (indirect)',
                    signal=biomass_signal.id,
                    biomass_signal=biomass_signal.id,
                    pre_smoothing=5,
                    post_smoothing=5).sort_values('Time')
        t = gr.Time.unique()
        gr_prof = gr.groupby('Time').mean().Rate.values
        self.growth_rate_prof = interp1d(t, gr_prof, bounds_error=False, fill_value='extrapolate')

        od = fj.analysis(media=media.id, 
                    strain=strain.id,
                    vector=vector.id,
                    type='Background Correct',
                    signal=biomass_signal.id,
                    biomass_signal=biomass_signal.id
                      ).sort_values('Time')
        t = od.Time.unique()
        biomass_prof = od.groupby('Time').mean().Measurement.values
        self.biomass_prof = interp1d(t, biomass_prof, bounds_error=False, fill_value='extrapolate')

    def biomass(self, t):
        return float(self.biomass_prof(t))

    def growth_rate(self, t):
        return float(self.growth_rate_prof(t))



import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import least_squares
from .receiver import *

class Nor:
    color = 'orange'
    shape = 's'
    def __init__(self, input, output, alpha, K, n, uri=None, sbol_comp=None):
        self.alpha = alpha
        self.K = K
        self.n = n
        self.input = input
        self.output = output
        self.uri = uri
        self.sbol_comp = sbol_comp

    def __str__(self):
        return 'NOR'

    def expression_rate(self, t, dt):
        input_repressor1 = self.input[0].concentration
        input_repressor2 = self.input[1].concentration
        r1 = (input_repressor1/self.K[0])**self.n[0]
        r2 = (input_repressor2/self.K[1])**self.n[1]
        r12 = (input_repressor1 * input_repressor2 / self.K[0] / self.K[1])**(self.n[0] + self.n[1])
        num = self.alpha[0] + self.alpha[1] * r1 + self.alpha[2] * r2 + self.alpha[3] * r12
        denom = 1 + r1 + r2 + r12
        return num / denom

    def forward_model(
        self,
        rep1_K=1, rep1_n=2,
        rep2_K=1, rep2_n=2,
        alpha0=1, alpha1=0, alpha2=0, alpha3=0,
        a_A=1e2, b_A=0, K_A=1, n_A=2,
        a_B=1e2, b_B=0, K_B=1, n_B=2,
        Dt=0.05,
        sim_steps=10,
        A=0, B=0,
        odval=[1]*100,
        gamma=0,
        rep1_0=0, rep2_0=0, fp_0=0,
        nt=100
    ):
        fp_list,A_list,B_list,t_list = [],[],[],[]
        rep1 = rep1_0
        rep2 = rep2_0
        fp = fp_0
        for t in range(nt):
            fp_list.append(fp)
            A_list.append(A)
            B_list.append(B)
            t_list.append(t * Dt)
            od = odval[t]
            for tt in range(sim_steps):
                time = (t + tt/sim_steps) * Dt
                # Repressor 1
                a = (A/K_A)**n_A
                nextrep1 = rep1 + (od * (a_A + b_A * a) /(1 + a) - gamma*rep1) * Dt/sim_steps

                # Repressor 2
                b = (B/K_B)**n_B
                nextrep2 = rep2 + (od * (a_B + b_B * b) /(1 + b) - gamma*rep2) * Dt/sim_steps

                # Reporter output
                r1 = (rep1/od/rep1_K)**rep1_n
                r2 = (rep2/od/rep2_K)**rep2_n
                r12 = (rep1*rep2/od/od/rep1_K/rep2_K)**(rep1_n+rep2_n)
                num = alpha0 + alpha1 * r1 + alpha2 * r2 + alpha3 * r12
                denom = 1 + r1 + r2 + r12
                expression_rate = num / denom
                nextfp = fp + ( od * expression_rate) * Dt/sim_steps

                # Update system
                rep1,rep2,fp = nextrep1,nextrep2,nextfp

        afp = np.array(fp_list).transpose()
        AA = np.array(A_list).transpose()
        BB = np.array(B_list).transpose()
        tt = np.array(t_list).transpose()
        t = np.arange(nt) * Dt
        return afp,AA,BB,tt

    def residuals(self, df, oddf, a_A, b_A, K_A, n_A, a_B, b_B, K_B, n_B, chem1, chem2, gamma): 
        def func(x): 
            rep1_K, rep1_n = x[0:2]
            rep2_K, rep2_n = x[2:4]
            alpha0, alpha1, alpha2, alpha3 = x[4:8]
            residual_list = []
            df_sorted = df.sort_values(['Sample', 'Time'])
            oddf_sorted = oddf.sort_values(['Sample', 'Time'])
            for samp_id,samp_data in df_sorted.groupby('Sample'):
                odval = oddf_sorted[oddf_sorted.Sample==samp_id].Measurement.values
                data = samp_data.Measurement.values
                p0_1 = 0
                p0_2 = data[0]
                # Get inducer concentrations if any
                chemical1 = samp_data.Chemical_id1.values[0]
                chemical2 = samp_data.Chemical_id2.values[0]
                if chemical1 == chem1:
                    A = samp_data.Concentration1.values[0]
                elif chemical2 == chem1:
                    A = samp_data.Concentration2.values[0]
                else:
                    A = 0
                if chemical1 == chem2:
                    B = samp_data.Concentration1.values[0]
                elif chemical2 == chem2:
                    B = samp_data.Concentration2.values[0]
                else:
                    B = 0
                t = samp_data.Time.values
                dt = np.mean(np.diff(t))
                nt = len(t)
                p,AA,BB,tt = self.forward_model(
                            rep1_K=rep1_K, rep1_n=rep1_n,
                            rep2_K=rep2_K, rep2_n=rep2_n,
                            alpha0=alpha0, alpha1=alpha1, alpha2=alpha2, alpha3=alpha3,
                            a_A=a_A, b_A=b_A, K_A=K_A, n_A=n_A,
                            a_B=a_B, b_B=b_B, K_B=K_B, n_B=n_B,
                            Dt=dt,
                            A=A, B=B,
                            odval=odval,
                            gamma=gamma,
                            rep1_0=0, rep2_0=0, fp_0=data[0],
                            nt=nt,
                        )
                model = p.ravel()
                residual = (data[1:] - model[1:]) 
                residual_list.append(residual) 
            return np.array(residual_list).ravel()
        return func

    def characterize(self, 
            flapjack, 
            receiver1,
            receiver2, 
            chemical1,
            chemical2,
            nor_inverter, 
            media, 
            strain, 
            signal, 
            biomass_signal,
            gamma
            ):
        # Get biomass time series
        biomass_df = flapjack.analysis(type='Background Correct', 
                            vector=nor_inverter,
                            media=media,
                            strain=strain,
                            signal=biomass_signal,
                            biomass_signal=biomass_signal
                         )
       
        # Characterize receiver1 Hill function
        rec1 = Receiver(None, None, 0, 0, 0, 0)
        rec1.characterize(
            flapjack,
            vector=receiver1,
            media=media,
            strain=strain,
            signal=signal,
            biomass_signal=biomass_signal
        )
        self.a_A = rec1.a
        self.b_A = rec1.b
        self.K_A = rec1.K
        self.n_A = rec1.n

        # Characterize receiver2 Hill function
        rec2 = Receiver(None, None, 0, 0, 0, 0)
        rec2.characterize(
            flapjack,
            vector=receiver2,
            media=media,
            strain=strain,
            signal=signal,
            biomass_signal=biomass_signal
        )
        self.a_B = rec2.a
        self.b_B = rec2.b
        self.K_B = rec2.K
        self.n_B = rec2.n

        # Characterize inverter
        nor_inverter_df = flapjack.analysis(type='Background Correct', 
                            vector=nor_inverter,
                            media=media,
                            strain=strain,
                            signal=signal,
                            biomass_signal=biomass_signal
                         )

        # Bounds for fitting
        lower_bounds = [0]*8
        upper_bounds = [1e8, 8, 1e8, 8, 1e8, 1e8, 1e8, 1e8]
        bounds = [lower_bounds, upper_bounds]

        '''
            rep1_K, rep1_n = x[0:2]
            rep2_K, rep2_n = x[2:4]
            alpha0, alpha1, alpha2, alpha3 = x[4:8]
        '''
        # Solve for parameters and profile
        residuals = self.residuals(
                                nor_inverter_df, biomass_df,
                                self.a_A, self.b_A, self.K_A, self.n_A,
                                self.a_B, self.b_B, self.K_B, self.n_B,
                                chemical1, chemical2,
                                gamma=gamma
                            )
        res = least_squares(residuals, [1,2,1,2,1,0,0,0], bounds=bounds)
        self.res = res
        self.rep1_K, self.rep1_n = res.x[0:2]
        self.rep2_K, self.rep2_n = res.x[2:4]
        self.alpha0, self.alpha1, self.alpha2, self.alpha3 = res.x[4:8]


from .assay import *
from .sample import *
from .genetic_network import *
from .metabolism import *
from .operators import *
from .geneproduct import *
from .supplement import *
from .util import *
from .colony import *



from setuptools import setup, find_packages


with open('README.md', 'r') as ld:
    long_description = ld.read()

setup(name='loica',
    version='1.0.1',
    author='Gonzalo Vidal',
    author_email='gsvidal@uc.cl',
    description='LOICA: Logical Operators for Integrated Cell Algorithms',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RudgeLab/LOICA',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        ],
    install_requires=[        
        'numpy', 
	    'networkx',
	    'pandas',
    	'sbol3',
    	'tyto',
   	    'tdqm'
        ],
    setup_requires=[
        'numpy',
	    'networkx',
	    'pandas',
    	'sbol3',
    	'tyto',
    	'tdqm'
        ],
    packages=find_packages(),
    python_requires='>=3'
)



import pickle
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import least_squares

def forward_model_growth(
    Dt=0.05,
    sim_steps=10,
    muval=[0]*100,
    od0=0,
    nt=100
):
    od_list, t_list = [],[]
    od = od0
    for t in range(nt):
        od_list.append(od)
        t_list.append([t * Dt])
        mu = muval[t]
        for tt in range(sim_steps):
            doddt = mu * od
            nextod = od + doddt * Dt/sim_steps
            od = nextod


    aod = np.array(od_list).transpose()
    tt = np.array(t_list).transpose()
    return aod,tt


def residuals_growth(data, epsilon, dt, t, n_gaussians): 
    def func(x): 
        od0 = x[0]
        muval = np.zeros_like(t)
        means = np.linspace(t.min(), t.max(), n_gaussians)
        vars = [(t.max()-t.min())/n_gaussians] * n_gaussians 
        heights = x[1:]
        for mean,var,height in zip(means, vars, heights):
            gaussian = height * np.exp(-(t-mean)*(t-mean) / var / 2) / np.sqrt(2 * np.pi * var)
            muval = muval + gaussian

        od,tt = forward_model_growth(
                    Dt=dt,
                    muval=muval,
                    od0=od0,
                    nt=len(t)
                )
        model = od
        residual = (data - model)  # / tt.ravel()[1:]
        tikhonov = heights
        result = np.concatenate((residual, epsilon * tikhonov))
        return result
    return func


def characterize_growth(flapjack, 
        vector, 
        media, 
        strain, 
        biomass_signal, 
        n_gaussians, 
        epsilon
        ):
    # Characterize growth rate profile
    biomass = flapjack.analysis(type='Background Correct', 
                        vector=vector,
                        media=media,
                        strain=strain,
                        signal=biomass_signal,
                        biomass_signal=biomass_signal
                     )
    biomass = biomass.sort_values('Time')
    t = biomass.Time.unique()
    dt = np.mean(np.diff(t))
    nt = len(t)

    biomass = biomass.groupby('Time').mean().Measurement.values
    lower_bounds = [0] + [0]*n_gaussians
    upper_bounds = [100] + [50]*n_gaussians
    bounds = [lower_bounds, upper_bounds]

    data = biomass
    res = least_squares(
            residuals_growth(data, epsilon=epsilon, dt=dt, t=t, n_gaussians=n_gaussians), 
            [0.01] + [1]*n_gaussians, 
            bounds=bounds
            )
    init_biomass = res.x[0]
    profile = np.zeros_like(t)
    means = np.linspace(t.min(), t.max(), n_gaussians)
    vars = [(t.max()-t.min())/n_gaussians] * n_gaussians 
    heights = res.x[1:]
    for mean,var,height in zip(means, vars, heights):
        gaussian = height * np.exp(-(t-mean)*(t-mean) / var / 2) / np.sqrt(2 * np.pi * var)
        profile = profile + gaussian
    mu_profile = interp1d(t, profile, fill_value='extrapolate', bounds_error=False)

    od,tt = forward_model_growth(
                    Dt=dt,
                    muval=profile,
                    od0=init_biomass,
                    nt=nt
                )
    return init_biomass, od, mu_profile

def save_loica(obj, filename):
    with open(f'{filename}.obj', 'wb') as outp:
        pickle.dump(obj, outp, pickle.HIGHEST_PROTOCOL)

def load_loica(filename):
    with open(f'{filename}.obj', 'rb') as inp:
        loica = pickle.load(inp)
    return loica